@@ -16,27 +16,31 @@
 #    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
 """Terminator by Chris Jones <cmsj@tenshu.net>"""
+import time, re, sys, os, platform
+
 import pygtk
 pygtk.require ("2.0")
 import gobject, gtk, pango
 
-#import version details
-from terminatorlib.version import *
+from terminatorlib.version import APP_NAME, APP_VERSION
 
-# import our configuration loader
 from terminatorlib import config
-from terminatorlib.config import dbg, err, debug
+from config import dbg, err, debug
 
-#import TerminatorTerm
+from terminatorlib.keybindings import TerminatorKeybindings
 from terminatorlib.terminatorterm import TerminatorTerm
 
 class TerminatorNotebookTabLabel(gtk.HBox):
-
   def __init__(self, title, notebook, terminator):
     gtk.HBox.__init__(self, False)
     self._notebook = notebook
     self.terminator = terminator
-    self._label = gtk.Label(title) 
+    self._label = gtk.Label(title)
+    tab_pos = notebook.get_tab_pos()
+    if tab_pos == gtk.POS_LEFT:
+      self._label.set_angle(90)
+    elif tab_pos == gtk.POS_RIGHT:
+      self._label.set_angle(270)
     icon = gtk.Image()
     icon.set_from_stock(gtk.STOCK_CLOSE, gtk.ICON_SIZE_MENU)
     self.pack_start(self._label, True, True)
@@ -50,8 +54,8 @@ class TerminatorNotebookTabLabel(gtk.HBox):
       self._button.connect('clicked', self.on_close)
       self._button.set_name("terminator-tab-close-button")
       self.connect("style-set", self.on_style_set)
-      
-      self._button.set_tooltip_text(_("Close Tab"))
+      if hasattr(self._button, "set_tooltip_text"): 
+        self._button.set_tooltip_text(_("Close Tab"))
       self.pack_start(self._button, False, False)
     self.show_all()
 
@@ -61,7 +65,6 @@ class TerminatorNotebookTabLabel(gtk.HBox):
 
   def on_close(self, widget):
     nbpages = self._notebook.get_n_pages()
-    print self._button.allocation.width, self._button.allocation.height
     for i in xrange(0,nbpages):
       if self._notebook.get_tab_label(self._notebook.get_nth_page(i)) == self:
         #dbg("[Close from tab] Found tab at position [%d]" % i)
@@ -79,42 +82,100 @@ class TerminatorNotebookTabLabel(gtk.HBox):
 
   def get_title(self):
     return self._label.get_text()
+  
+  def height_request(self):
+    return self.size_request()[1]
+
+  def width_request(self):
+    return self.size_request()[0]
+
 
 class Terminator:
-  def __init__ (self, profile = None, command = None, fullscreen = False, maximise = False, borderless = False):
+  def __init__ (self, profile = None, command = None, fullscreen = False,
+                maximise = False, borderless = False, no_gconf = False,
+                geometry = None):
     self.profile = profile
     self.command = command
 
     self._zoomed = False
     self._maximised = False
     self._fullscreen = False
-    self._f11_modifier = False
+    self._geometry = geometry
+    self.debugaddress = None
+    self.start_cwd = os.getcwd()
     self.term_list = []
     stores = []
     stores.append (config.TerminatorConfValuestoreRC ())
+    
+    self._tab_reorderable = True
+    if not hasattr(gtk.Notebook, "set_tab_reorderable") or not hasattr(gtk.Notebook, "get_tab_reorderable"):
+      self._tab_reorderable = False
 
-    try:
-      import gconf
-      if self.profile:
-        self.profile = gconf.escape_key (self.profile, -1)
-      store = config.TerminatorConfValuestoreGConf (self.profile)
-      store.set_reconfigure_callback (self.reconfigure_vtes)
-      dbg ('Terminator__init__: comparing %s and %s'%(self.profile, store.profile.split ('/').pop ()))
-      if self.profile == store.profile.split ('/').pop ():
-        # If we have been given a profile, and we loaded it, we should be higher priority than RC
-        dbg ('Terminator__init__: placing GConf before RC')
-        stores.insert (0, store)
-      else:
-        stores.append (store)
-    except:
-      pass
+    if not no_gconf:
+      try:
+        import gconf
+        if self.profile:
+          self.profile = gconf.escape_key (self.profile, -1)
+        store = config.TerminatorConfValuestoreGConf (self.profile)
+        store.set_reconfigure_callback (self.reconfigure_vtes)
+        dbg ('Terminator__init__: comparing %s and %s'%(self.profile, store.profile.split ('/').pop ()))
+        if self.profile == store.profile.split ('/').pop ():
+          # If we have been given a profile, and we loaded it, we should be higher priority than RC
+          dbg ('Terminator__init__: placing GConf before RC')
+          stores.insert (0, store)
+        else:
+          stores.append (store)
+      except Exception, e:
+        # This should probably be ImportError; what else might it throw?
+        dbg("GConf setup threw exception %s" % str(e))
 
     self.conf = config.TerminatorConfig (stores)
 
+    # Sort out cwd detection code, if available
+    self.pid_get_cwd = lambda pid: None
+    if platform.system() == 'FreeBSD':
+      try:
+        from terminatorlib import freebsd
+        self.pid_get_cwd = freebsd.get_process_cwd
+        dbg ('Using FreeBSD self.pid_get_cwd')
+      except (OSError, NotImplementedError, ImportError):
+        dbg ('FreeBSD version too old for self.pid_get_cwd')
+    elif platform.system() == 'Linux':
+      dbg ('Using Linux self.pid_get_cwd')
+      self.pid_get_cwd = lambda pid: os.path.realpath ('/proc/%s/cwd' % pid)
+    else:
+      dbg ('Unable to set a self.pid_get_cwd, unknown system: %s' % platform.system)
+
+    # import a library for viewing URLs
+    try:
+      dbg ('Trying to import gnome for X session and backup URL handling support')
+      global gnome
+      import gnome, gnome.ui
+      self.gnome_program = gnome.init(APP_NAME, APP_VERSION)
+      self.url_show = gnome.url_show
+      
+      # X session saving support
+      self.gnome_client = gnome.ui.master_client()
+      self.gnome_client.connect_to_session_manager()
+      self.gnome_client.connect('save-yourself', self.save_yourself)
+      self.gnome_client.connect('die', self.die)
+    except ImportError:
+      # webbrowser.open() is not really useful, but will do as a fallback
+      dbg ('gnome not available, no X session support, backup URL handling via webbrowser module')
+      import webbrowser
+      self.url_show = webbrowser.open
+
+
     self.icon_theme = gtk.IconTheme ()
 
+    self.keybindings = TerminatorKeybindings()
     if self.conf.f11_modifier:
-      self._f11_modifier = True
+      config.Defaults['keybindings']['full_screen'] = '<Ctrl><Shift>F11'
+      print "Warning: Config setting f11_modifier is deprecated and will be removed in version 1.0"
+      print "Please add the following to the end of your terminator config:"
+      print "[keybindings]"
+      print "full_screen = <Ctrl><Shift>F11"
+    self.keybindings.configure(self.conf.keybindings)
 
     if self.conf.handle_size in xrange (0,6):
       gtk.rc_parse_string("""
@@ -138,6 +199,11 @@ class Terminator:
     self.window = gtk.Window ()
     self.window.set_title (APP_NAME.capitalize())
 
+    if self._geometry is not None:
+      dbg("Geometry=%s" % self._geometry)
+      if not self.window.parse_geometry(self._geometry):
+        err(_("Invalid geometry string %s") % repr(self._geometry))
+
     try:
       self.window.set_icon (self.icon_theme.load_icon (APP_NAME, 48, 0))
     except:
@@ -162,9 +228,6 @@ class Terminator:
 
     # Set RGBA colormap if possible so VTE can use real alpha
     # channels for transparency.
-    #TODO maybe it work in every case
-    # LP#242852 real transparency causes graphical glitches,
-    # real transparency can be disabled in ~/.config/terminator/config
     if self.conf.enable_real_transparency:
       screen = self.window.get_screen()
       colormap = screen.get_rgba_colormap()
@@ -180,6 +243,55 @@ class Terminator:
     term._titlebox.hide()
     self.window.show ()
     term.spawn_child ()
+    self.save_yourself ()
+
+  def die(self, *args):
+    gtk.main_quit ()
+
+  def save_yourself (self, *args):
+    """ Save as much of our state as possible for the X session manager """
+    dbg("Saving session for xsm")
+    args = [sys.argv[0],
+        ("--geometry=%dx%d" % self.window.get_size()) + ("+%d+%d" % self.window.get_position())]
+
+    # OptionParser should really help us out here
+    drop_next_arg = False
+    geompatt = re.compile(r'^--geometry(=.+)?')
+    for arg in sys.argv[1:]:
+      mo = geompatt.match(arg)
+      if mo:
+        if not mo.group(1):
+          drop_next_arg = True
+      elif not drop_next_arg and arg not in ('--maximise', '-m', '--fullscreen', '-f'):
+        args.append(arg)
+        drop_next_arg = False
+
+    if self._maximised:
+      args.append('--maximise')
+
+    if self._fullscreen:
+      args.append('--fullscreen')
+
+    # We can't set an interpreter because Gnome unconditionally spams it with
+    # --sm-foo-bar arguments before our own argument list. *mutter*
+    # So, hopefully your #! line is correct.  If not, we could write out
+    # a shell script with the interpreter name etc.
+    c = self.gnome_client
+    c.set_program(sys.argv[0])
+    dbg("Session restart command: %s with args %s in %s" % (sys.argv[0], repr(args), self.start_cwd))
+
+    c.set_restart_style(gnome.ui.RESTART_IF_RUNNING)
+    c.set_current_directory(self.start_cwd)
+    try:
+      c.set_restart_command(args)
+      c.set_clone_command(args)
+    except (TypeError,AttributeError):
+      # Apparantly needed for some Fedora systems
+      # see http://trac.nicfit.net/mesk/ticket/137
+      dbg("Gnome bindings have weird set_clone/restart_command")
+      c.set_restart_command(len(args), args)
+      c.set_clone_command(len(args), args)
+    return True
 
   def maximize (self):
     """ Maximize the Terminator window."""
@@ -196,9 +308,9 @@ class Terminator:
       self.window.fullscreen ()
 
   def on_window_state_changed (self, window, event):
-    state = event.new_window_state & gtk.gdk.WINDOW_STATE_FULLSCREEN
-    self._fullscreen = bool (state)
-
+    self._fullscreen = bool (event.new_window_state & gtk.gdk.WINDOW_STATE_FULLSCREEN)
+    self._maximised = bool (event.new_window_state & gtk.gdk.WINDOW_STATE_MAXIMIZED)
+    dbg("window state changed: fullscreen: %s, maximised: %s" % (self._fullscreen, self._maximised))
     return (False)
 
   def on_delete_event (self, window, event, data=None):
@@ -207,12 +319,13 @@ class Terminator:
 
     # show dialog
     dialog = gtk.Dialog (_("Close?"), window, gtk.DIALOG_MODAL)
-    cancel = dialog.add_button(gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT)
-    close_all = dialog.add_button(gtk.STOCK_CLOSE, gtk.RESPONSE_ACCEPT)
-    label = close_all.get_children()[0].get_children()[0].get_children()[1].set_label(_("Close _All Terminals"))
     dialog.set_has_separator (False)
     dialog.set_resizable (False)
 
+    cancel = dialog.add_button(gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT)
+    close_all = dialog.add_button(gtk.STOCK_CLOSE, gtk.RESPONSE_ACCEPT)
+    label = close_all.get_children()[0].get_children()[0].get_children()[1].set_label(_("Close All _Terminals"))
+
     primairy = gtk.Label (_('<big><b>Close all terminals?</b></big>'))
     primairy.set_use_markup (True)
     primairy.set_alignment (0, 0.5)
@@ -238,7 +351,7 @@ class Terminator:
     return not (result == gtk.RESPONSE_ACCEPT)
 
   def on_destroy_event (self, widget, data=None):
-    gtk.main_quit ()
+    self.die()
 
   # keybindings for the whole terminal window (affects the main
   # windows containing the splited terminals)
@@ -248,15 +361,14 @@ class Terminator:
           * F11:              toggle fullscreen state of the window.
           * CTRL - SHIFT - Q: close all terminals
     """
-    keyname = gtk.gdk.keyval_name (event.keyval)
-    mask = gtk.gdk.CONTROL_MASK | gtk.gdk.SHIFT_MASK
+    mapping = self.keybindings.lookup(event)
 
-    if (keyname == 'F11' and (self._f11_modifier == False or event.state & mask)):
-      self.fullscreen_toggle ()
-      return (True)
-
-    if (event.state & mask) == mask:
-      if keyname == 'Q':
+    if mapping:
+      dbg("on_key_press: lookup found %s" % repr(mapping))
+      if mapping == 'full_screen':
+        self.fullscreen_toggle ()
+        return (True)
+      elif mapping == 'close_window':
         if not self.on_delete_event (window, gtk.gdk.Event (gtk.gdk.DELETE)):
           self.on_destroy_event (window, gtk.gdk.Event (gtk.gdk.DESTROY))
 
@@ -303,6 +415,7 @@ class Terminator:
       #not the only term in the notebook/window anymore, need to reshow the title
       dbg ('SEGBUG: Showing _titlebox')
       widget._titlebox.show()
+      terminal._titlebox.show()
       
     if isinstance (parent, gtk.Notebook):
       dbg ('SEGBUG: Parent is a notebook')
@@ -323,12 +436,13 @@ class Terminator:
       #parent.remove_page(page)
       pane.show()
       parent.insert_page(pane, None, page)
-      notebooktablabel = TerminatorNotebookTabLabel(widget._vte.get_window_title(), parent, self)
+      notebooktablabel = TerminatorNotebookTabLabel(widget.get_window_title(), parent, self)
       parent.set_tab_label(pane,notebooktablabel)
       parent.set_tab_label_packing(pane, True, True, gtk.PACK_START)
-      parent.set_tab_reorderable(pane, True)
+      if self._tab_reorderable:
+        parent.set_tab_reorderable(pane, True)
       parent.set_current_page(page)
-      
+
       position = (vertical) and parent.allocation.height \
                              or parent.allocation.width
 
@@ -443,13 +557,13 @@ class Terminator:
     dbg("[ERROR] unsupported class %s in _notebook_last_term" % child.__class__.__name__)
     return None
   
-  def newtab(self,widget, toplevel = False):
+  def newtab(self,widget, toplevel = False, command = None):
     if self._zoomed:
       # We don't want to add a new tab while we are zoomed in on a terminal
       dbg ("newtab function called, but Terminator was in zoomed terminal mode.")
       return
 
-    terminal = TerminatorTerm (self, self.profile, None, widget.get_cwd())
+    terminal = TerminatorTerm (self, self.profile, command, widget.get_cwd())
     #only one term, we don't show the title
     terminal._titlebox.hide()
     if self.conf.extreme_tabs and not toplevel:
@@ -464,10 +578,13 @@ class Terminator:
        ((self.conf.extreme_tabs and not toplevel) or not isinstance(child, gtk.Notebook))):
       #no notebook yet.
       notebook = gtk.Notebook()
-      #notebook.set_tab_pos(gtk.POS_TOP)
-      notebook.connect('page-reordered',self.on_page_reordered)
+      if self._tab_reorderable:
+        notebook.connect('page-reordered',self.on_page_reordered)
+        notebook.set_tab_reorderable(widget, True)
       notebook.set_property('homogeneous', True)
-      notebook.set_tab_reorderable(widget, True)
+      # Config validates this.
+      pos = getattr(gtk, "POS_%s" % self.conf.tab_position.upper())
+      notebook.set_tab_pos(pos)
       
       if isinstance(parent, gtk.Paned):
         if parent.get_child1() == child:
@@ -479,16 +596,23 @@ class Terminator:
       elif isinstance(parent, gtk.Window):
          child.reparent(notebook)
          parent.add(notebook)
-      notebook.set_tab_reorderable(child,True)
+      if self._tab_reorderable:
+        notebook.set_tab_reorderable(child,True)
       notebooklabel = ""
       if isinstance(child, TerminatorTerm):
         child._titlebox.hide()
-      if widget._vte.get_window_title() is not None:
-        notebooklabel = widget._vte.get_window_title()
+      if widget.get_window_title() is not None:
+        notebooklabel = widget.get_window_title()
       notebooktablabel = TerminatorNotebookTabLabel(notebooklabel, notebook, self)
       notebook.set_tab_label(child, notebooktablabel)
       notebook.set_tab_label_packing(child, True, True, gtk.PACK_START)
-      notebook.show()  
+
+      wal = self.window.allocation
+      if not (self._maximised or self._fullscreen):
+        self.window.resize(wal.width,
+            min(wal.height + notebooktablabel.height_request(), gtk.gdk.screen_height()))
+
+      notebook.show()
     elif isinstance(parent, gtk.Notebook):
       notebook = parent
     elif isinstance(parent, gtk.Window) and isinstance(child, gtk.Notebook):
@@ -503,17 +627,12 @@ class Terminator:
     notebook.append_page(terminal,None)
     terminal.show ()
     terminal.spawn_child ()
-    ## Some gtk/vte weirdness
-    ## If we don't use this silly test,
-    ## terminal._vte.get_window_title() might return
-    ## bogus values
-    notebooklabel = ""
-    if terminal._vte.get_window_title() is not None:
-      notebooklabel = terminal._vte.get_window_title()
+    notebooklabel = terminal.get_window_title()
     notebooktablabel = TerminatorNotebookTabLabel(notebooklabel, notebook, self)
     notebook.set_tab_label(terminal, notebooktablabel)
     notebook.set_tab_label_packing(terminal, True, True, gtk.PACK_START)
-    notebook.set_tab_reorderable(terminal,True)
+    if self._tab_reorderable:
+      notebook.set_tab_reorderable(terminal,True)
     ## Now, we set focus on the new term
     notebook.set_current_page(-1)
     terminal._vte.grab_focus ()
@@ -526,7 +645,7 @@ class Terminator:
     self.term_list.insert (index + 1, terminal)
     return (True)
   
-  def splitaxis (self, widget, vertical=True):
+  def splitaxis (self, widget, vertical=True, command=None):
     """ Split the provided widget on the horizontal or vertical axis. """
     if self._zoomed:
       # We don't want to split the terminal while we are in zoomed mode
@@ -535,7 +654,7 @@ class Terminator:
 
     # create a new terminal and parent pane.
     dbg ('SEGBUG: Creating TerminatorTerm')
-    terminal = TerminatorTerm (self, self.profile, None, widget.get_cwd())
+    terminal = TerminatorTerm (self, self.profile, command, widget.get_cwd())
     dbg ('SEGBUG: Created TerminatorTerm')
     pos = vertical and "right" or "bottom"
     dbg ('SEGBUG: Position is: %s'%pos)
@@ -587,7 +706,8 @@ class Terminator:
         grandparent.insert_page(sibling, None,page)
         grandparent.set_tab_label(sibling, TerminatorNotebookTabLabel("",grandparent, self))
         grandparent.set_tab_label_packing(sibling, True, True, gtk.PACK_START)
-        grandparent.set_tab_reorderable(sibling, True)
+        if self._tab_reorderable:
+          grandparent.set_tab_reorderable(sibling, True)
         grandparent.set_current_page(page)
       else:
         grandparent.remove (parent)
@@ -606,6 +726,8 @@ class Terminator:
       index = self.term_list.index (widget)
       self.term_list.remove (widget)
       if nbpages == 1:
+        if self.window.allocation.height != gtk.gdk.screen_height():
+          self.window.resize(self.window.allocation.width, min(self.window.allocation.height - parent.get_tab_label(parent.get_nth_page(0)).height_request(), gtk.gdk.screen_height()))
         sibling = parent.get_nth_page(0)
         parent.remove(sibling)
         gdparent = parent.get_parent()
@@ -787,7 +909,7 @@ class Terminator:
     return self.get_first_parent_paned(parent, vertical)
 
   def get_first_notebook_page(self, widget):
-    if isinstance (widget, gtk.Window):
+    if isinstance (widget, gtk.Window) or widget is None:
       return None
     parent = widget.get_parent()
     if isinstance (parent, gtk.Notebook):
@@ -817,6 +939,7 @@ class Terminator:
           (isinstance(widget, TerminatorTerm) and isinstance(widget.get_parent(),gtk.Paned))\
           :
         widget._titlebox.show()
+
     widget._vte.grab_focus()
 
   def zoom_term (self, widget, fontscale = False):
@@ -892,9 +1015,16 @@ class Terminator:
 
       self.window.remove(widget)
       self.window.add(self.window_child)
-      self.old_parent.add(widget)
       if isinstance(self.old_parent, gtk.Notebook):
+        self.old_parent.insert_page(widget, None, self.old_page)
+        self.old_parent.set_tab_label(widget, TerminatorNotebookTabLabel("", self.old_parent, self))
+        self.old_parent.set_tab_label_packing(widget, True, True, gtk.PACK_START)
+        if self._tab_reorderable:
+          self.old_parent.set_tab_reorderable(widget, True)
         self.old_parent.set_current_page(self.old_page)
 
+      else:
+        self.old_parent.add(widget)
+
       widget._vte.grab_focus ()
 
