@@ -19,7 +19,7 @@
 import pygtk
 pygtk.require ("2.0")
 import gobject, gtk, pango
-import os, platform, sys
+import os, sys, subprocess, pwd
 
 #import version details
 from terminatorlib.version import *
@@ -34,7 +34,7 @@ from terminatorlib.encoding import TerminatorEncoding
 # import vte-bindings
 try:
   import vte
-except:
+except ImportError:
   error = gtk.MessageDialog (None, gtk.DIALOG_MODAL, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK,
     _('You need to install python bindings for libvte ("python-vte" in debian/ubuntu)'))
   error.run()
@@ -44,6 +44,7 @@ class TerminatorTerm (gtk.VBox):
 
   matches = {}
   TARGET_TYPE_VTE = 8
+  _custom_font_size = None
 
   def __init__ (self, terminator, profile = None, command = None, cwd = None):
     gtk.VBox.__init__ (self)
@@ -51,35 +52,6 @@ class TerminatorTerm (gtk.VBox):
     self.conf = terminator.conf
     self.command = command
 
-    # Sort out cwd detection code, if available
-    self.pid_get_cwd = lambda pid: None
-    if platform.system() == 'FreeBSD':
-      try:
-        from terminatorlib import freebsd
-        self.pid_get_cwd = lambda pid: freebsd.get_process_cwd(pid)
-        dbg ('Using FreeBSD self.pid_get_cwd')
-      except:
-        dbg ('FreeBSD version too old for self.pid_get_cwd')
-        pass
-    elif platform.system() == 'Linux':
-      dbg ('Using Linux self.pid_get_cwd')
-      self.pid_get_cwd = lambda pid: os.path.realpath ('/proc/%s/cwd' % pid)
-    else:
-      dbg ('Unable to set a self.pid_get_cwd, unknown system: %s'%platform.system)
-
-    # import a library for viewing URLs
-    try:
-      # gnome.url_show() is really useful
-      dbg ('url_show: importing gnome module')
-      import gnome
-      gnome.init ('terminator', 'terminator')
-      self.url_show = gnome.url_show
-    except:
-      # webbrowser.open() is not really useful, but will do as a fallback
-      dbg ('url_show: gnome module failed, using webbrowser')
-      import webbrowser
-      self.url_show = webbrowser.open
-
     self.cwd = cwd or os.getcwd();
     if not os.path.exists(self.cwd) or not os.path.isdir(self.cwd):
       self.cwd = pwd.getpwuid(os.getuid ())[5]
@@ -87,9 +59,14 @@ class TerminatorTerm (gtk.VBox):
     self.clipboard = gtk.clipboard_get (gtk.gdk.SELECTION_CLIPBOARD)
     self.scrollbar_position = self.conf.scrollbar_position
 
+    self._composited_support = True
     self._vte = vte.Terminal ()
+    if not hasattr(self._vte, "set_opacity") or not hasattr(self._vte, "is_composited"):
+      self._composited_support = False
+    #self._vte.set_double_buffered(True)
     self._vte.set_size (80, 24)
     self.reconfigure_vte ()
+    self._vte._expose_data = None
     self._vte.show ()
 
     self._termbox = gtk.HBox ()
@@ -98,9 +75,30 @@ class TerminatorTerm (gtk.VBox):
     self._title.show()
     self._titlebox =  gtk.EventBox()
     self._titlebox.add(self._title)
+
+    self._search_string = None
+    self._searchbox = gtk.HBox()
+    self._searchinput = gtk.Entry()
+    self._searchinput.set_activates_default(True)
+    self._searchinput.show()
+
+    self._searchinput.connect('activate', self.do_search)
+    self._searchinput.connect('key-press-event', self.search_keypress)
+
+    slabel = gtk.Label()
+    slabel.set_text("Finding text:")
+    slabel.show()
+    self._searchbox.pack_start(slabel, False)
+    self._search_result_label = gtk.Label()
+    self._search_result_label.set_text("")
+    self._search_result_label.show()
+    self._searchbox.pack_end(self._search_result_label, False)
+    self._searchbox.pack_start(self._searchinput)
+
     self.show()
     self.pack_start(self._titlebox, False)
     self.pack_start(self._termbox)
+    self.pack_end(self._searchbox)
 
     if self.conf.titlebars:
       self._titlebox.show()
@@ -121,8 +119,7 @@ class TerminatorTerm (gtk.VBox):
 
     self._vte.connect ("key-press-event", self.on_vte_key_press)
     self._vte.connect ("button-press-event", self.on_vte_button_press)
-    self._vte.connect ("popup-menu", self.on_vte_popup_menu)
-    
+    self._vte.connect ("popup-menu", self.create_popup_menu)
     """drag and drop"""
     srcvtetargets = [ ( "vte", gtk.TARGET_SAME_APP, self.TARGET_TYPE_VTE ) ]
     dsttargets = [ ( "vte", gtk.TARGET_SAME_APP, self.TARGET_TYPE_VTE ), ('text/plain', 0, 0) , ("STRING", 0, 0), ("COMPOUND_TEXT", 0, 0)]
@@ -140,12 +137,13 @@ class TerminatorTerm (gtk.VBox):
 
     if self.conf.copy_on_selection:
       self._vte.connect ("selection-changed", lambda widget: self._vte.copy_clipboard ())
-
-    self._vte.connect ("composited-changed", self.on_composited_changed)
+    if self._composited_support :
+      self._vte.connect ("composited-changed", self.on_composited_changed)
     self._vte.connect ("window-title-changed", self.on_vte_title_change)
     self._vte.connect ("grab-focus", self.on_vte_focus)
     self._vte.connect ("focus-out-event", self.on_vte_focus_out)
     self._vte.connect ("focus-in-event", self.on_vte_focus_in)
+    self._vte.connect ("resize-window", self.on_resize_window)
 
     exit_action = self.conf.exit_action
     if exit_action == "restart":
@@ -158,7 +156,7 @@ class TerminatorTerm (gtk.VBox):
     self._vte.add_events (gtk.gdk.ENTER_NOTIFY_MASK)
     self._vte.connect ("enter_notify_event", self.on_vte_notify_enter)
 
-    self.add_matches()
+    self.add_matches(posix = self.conf.try_posix_regexp)
 
     dbg ('SEGBUG: Setting http_proxy')
     env_proxy = os.getenv ('http_proxy')
@@ -172,25 +170,50 @@ class TerminatorTerm (gtk.VBox):
   def openurl (self, url):
     dbg ('openurl: viewing %s'%url)
     try:
-      if subprocess.call(["xdg-open", url]) != 0:
-        dbg ('openurl: xdg-open failed')
-        raise
+      dbg ('openurl: calling xdg-open')
+      subprocess.Popen(["xdg-open", url])
     except:
+      dbg ('openurl: xdg-open failed')
       try:
         dbg ('openurl: calling url_show')
-        self.url_show (url)
+        self.terminator.url_show (url)
       except:
         dbg ('openurl: url_show failed. No URL for you')
         pass
- 
+
+  def on_resize_window(self, widget, width, height):
+    dbg ('Resize window triggered on %s: %dx%d' % (widget, width, height))
+
   def on_drag_begin(self, widget, drag_context, data):
     dbg ('Drag begins')
     widget.drag_source_set_icon_pixbuf(self.terminator.icon_theme.load_icon (APP_NAME, 48, 0))
-    
+
   def on_drag_data_get(self,widget, drag_context, selection_data, info, time, data):
     dbg ("Drag data get")
     selection_data.set("vte",info, str(data.terminator.term_list.index (self)))
 
+  def on_expose_event(self, widget, event):
+    if widget._expose_data is None:
+      return False
+
+    color = widget._expose_data['color']
+    coord = widget._expose_data['coord']
+    
+    context = widget.window.cairo_create()
+    #leaving those xxx_group as they could be usefull
+    ##http://macslow.thepimp.net/?p=153
+    #context.push_group()
+    context.set_source_rgba(color.red, color.green, color.blue, 0.5)
+    if len(coord) > 0 :
+      context.move_to(coord[len(coord)-1][0],coord[len(coord)-1][1])
+      for i in coord:
+        context.line_to(i[0],i[1])
+    
+    context.fill()
+    #context.pop_group_to_source()
+    #context.paint()
+    return False
+
   def on_drag_motion(self, widget, drag_context, x, y, time, data): 
     dbg ("Drag Motion on ")
     """
@@ -215,16 +238,11 @@ text/plain
 
     alloc = widget.allocation
     rect = gtk.gdk.Rectangle(0, 0, alloc.width, alloc.height)
-    widget.window.invalidate_rect(rect, True)
-    widget.window.process_updates(True)
     
-    context = widget.window.cairo_create()
     if self.conf.use_theme_colors:
       color = self._vte.get_style ().text[gtk.STATE_NORMAL]
     else:
       color = gtk.gdk.color_parse (self.conf.foreground_color)
-
-    context.set_source_rgba(color.red, color.green, color.blue, 0.5)
      
     pos = self.get_location(widget, x, y)
     topleft = (0,0)
@@ -246,18 +264,22 @@ text/plain
       coord = (topleft, topmiddle, bottommiddle, bottomleft)
     if pos == "bottom":
       coord = (bottomleft, bottomright, middleright , middleleft) 
-     
-    if len(coord) > 0 :
-      context.move_to(coord[len(coord)-1][0],coord[len(coord)-1][1])
-      for i in coord:
-        context.line_to(i[0],i[1])
-      
-      context.fill()
- 
+    
+    
+    #here, we define some widget internal values
+    widget._expose_data = { 'color': color, 'coord' : coord }
+    #redraw by forcing an event
+    connec = widget.connect('expose-event', self.on_expose_event)
+    widget.window.invalidate_rect(rect, True)
+    widget.window.process_updates(True)
+    #finaly reset the values
+    widget.disconnect(connec)
+    widget._expose_data = None
+
   def on_drag_drop(self, widget, drag_context, x, y, time):
     parent = widget.get_parent()
     dbg ('Drag drop on %s'%parent)
-    
+
   def on_drag_data_received(self, widget, drag_context, x, y, selection_data, info, time, data):
     dbg ("Drag Data Received")
     if selection_data.type == 'text/plain':
@@ -318,7 +340,7 @@ text/plain
       pos = "bottom"
     return pos
 
-  def add_matches (self, lboundry="[[:<:]]", rboundry="[[:>:]]"):
+  def add_matches (self, posix = True):
     userchars = "-A-Za-z0-9"
     passchars = "-A-Za-z0-9,?;.:/!%$^*&~\"#'"
     hostchars = "-A-Za-z0-9"
@@ -326,88 +348,122 @@ text/plain
     schemes   = "(news:|telnet:|nntp:|file:/|https?:|ftps?:|webcal:)"
     user      = "[" + userchars + "]+(:[" + passchars + "]+)?"
     urlpath   = "/[" + pathchars + "]*[^]'.}>) \t\r\n,\\\"]"
-    
+
+    if posix:
+      dbg ('add_matches: Trying POSIX URL regexps.  Set try_posix_regexp = False in config to only try GNU if you get (harmless) VTE warnings.')
+      lboundry = "[[:<:]]"
+      rboundry = "[[:>:]]"
+    else: # GNU
+      dbg ('add_matches: Trying GNU URL regexps.  Set try_posix_regexp = True in config if URLs are not detected.')
+      lboundry = "\\<"
+      rboundry = "\\>"
+
     self.matches['full_uri'] = self._vte.match_add(lboundry + schemes + "//(" + user + "@)?[" + hostchars  +".]+(:[0-9]+)?(" + urlpath + ")?" + rboundry + "/?")
 
-    # FreeBSD works with [[:<:]], Linux works with \<
     if self.matches['full_uri'] == -1:
-      if lboundry != "\\<":
-        self.add_matches(lboundry = "\\<", rboundry = "\\>")
+      if posix:
+        err ('add_matches: POSIX match failed, trying GNU')
+        self.add_matches(posix = False)
+      else:
+        err ('add_matches: Failed adding URL match patterns')
     else:
       self.matches['addr_only'] = self._vte.match_add (lboundry + "(www|ftp)[" + hostchars + "]*\.[" + hostchars + ".]+(:[0-9]+)?(" + urlpath + ")?" + rboundry + "/?")
       self.matches['email'] = self._vte.match_add (lboundry + "(mailto:)?[a-zA-Z0-9][a-zA-Z0-9.+-]*@[a-zA-Z0-9][a-zA-Z0-9-]*\.[a-zA-Z0-9][a-zA-Z0-9-]+[.a-zA-Z0-9-]*" + rboundry)
       self.matches['nntp'] = self._vte.match_add (lboundry + '''news:[-A-Z\^_a-z{|}~!"#$%&'()*+,./0-9;:=?`]+@[-A-Za-z0-9.]+(:[0-9]+)?''' + rboundry)
 
+  def _path_lookup(self, command):
+    if os.path.isabs (command):
+      if os.path.isfile (command):
+        return command
+      else:
+        return None
+    elif command[:2] == './' and os.path.isfile(command):
+      dbg('path_lookup: Relative filename "%s" found in cwd' % command)
+      return command
+
+    try:
+      paths = os.environ['PATH'].split(':')
+      if len(paths[0]) == 0: raise (ValueError)
+    except (ValueError, NameError):
+      dbg('path_lookup: PATH not set in environment, using fallbacks')
+      paths = ['/usr/local/bin', '/usr/bin', '/bin']
+
+    dbg('path_lookup: Using %d paths: %s' % (len(paths), paths))
+
+    for path in paths:
+      target = os.path.join (path, command)
+      if os.path.isfile (target):
+        dbg('path_lookup: found "%s"' % target)
+        return target
+
+    dbg('path_lookup: Unable to locate "%s"' % command)
+
+  def _shell_lookup(self):
+    shells = [os.getenv('SHELL'), pwd.getpwuid(os.getuid())[6],
+              'bash', 'zsh', 'tcsh', 'ksh', 'csh', 'sh']
+
+    for shell in shells:
+      if shell is None: continue
+      elif os.path.isfile (shell):
+        return shell
+      else:
+        rshell = self._path_lookup(shell)
+        if rshell is not None:
+          dbg('shell_lookup: Found "%s" at "%s"' % (shell, rshell))
+          return rshell
+
+    dbg('shell_lookup: Unable to locate a shell')
+
   def spawn_child (self, event=None):
     update_records = self.conf.update_records
     login = self.conf.login_shell
     args = []
-    shell = ''
+    shell = None
+    command = None
 
     if self.command:
-      dbg ('spawn_child: using self.command: %s'%self.command)
-      args = self.command
-      shell = self.command[0]
+      dbg ('spawn_child: using self.command: %s' % self.command)
+      command = self.command
     elif self.conf.use_custom_command:
-      dbg ('spawn_child: using custom command: %s'%self.conf.custom_command)
-      args = self.conf.custom_command.split ()
-      shell = args[0]
+      dbg ('spawn_child: using custom command: %s' % self.conf.custom_command)
+      command = self.conf.custom_command
 
-    try:
-      if os.environ['PATH'] == "":
-        raise (ValueError)
-      paths = os.environ['PATH'].split(':')
-    except:
-      paths = ['/usr/local/bin', '/usr/bin', '/bin']
-    dbg ('spawn_child: found paths: "%s"'%paths)
+    if type(command) is list:
+      # List of arguments from -x
+      dbg('spawn_child: Bypassing shell and trying to run "%s" directly' % command[0])
+      shell = self._path_lookup(command[0])
+      args = command
+    else:
+      shell = self._shell_lookup()
 
-    if self.conf.use_custom_command and shell[0] != '/':
-     for path in paths:
-       dbg ('spawn_child: looking for pathless custom command "%s"'%os.path.join (path, shell))
-       if os.path.exists (os.path.join (path, shell)):
-         shell = os.path.join (path, shell)
-         break
+      if self.conf.login_shell:
+        args.insert(0, "-%s" % shell)
+      else:
+        args.insert(0, shell)
 
-    if not self.command and not os.path.exists (shell):
-      dbg ('spawn_child: hunting for a command')
-      shell = os.getenv ('SHELL') or ''
-      args = []
-      if not os.path.exists (shell):
-        dbg ('spawn_child: No usable shell in $SHELL (%s)'%os.getenv('SHELL'))
-        shell = pwd.getpwuid (os.getuid ())[6] or ''
-        if not os.path.exists (shell):
-          for i in ['bash','zsh','tcsh','ksh','csh','sh']:
-            for p in paths:
-              shell = os.path.join(p, i)
-              dbg ('spawn_child: Checking if "%s" exists'%shell)
-              if not os.path.exists (shell):
-                dbg ('spawn_child: %s does not exist'%shell)
-                continue
-              else:
-                dbg ('spawn_child: %s does exist'%shell)
-                break
-            if os.path.exists (shell):
-              break
+      if command is not None:
+        args += ['-c', command]
 
-    if not self.command and not os.path.exists (shell):
+    if shell is None:
       # Give up, we're completely stuck
       err (_('Unable to find a shell'))
       gobject.timeout_add (100, self.terminator.closeterm, self)
       return (-1)
 
-    if not args:
-      args.append (shell)
-
-    if self.conf.login_shell:
-      args[0] = "-%s"%args[0]
-
     dbg ('SEGBUG: Setting WINDOWID')
-    os.putenv ('WINDOWID', '%s'%self._vte.get_parent_window().xid)
+    os.putenv ('WINDOWID', '%s' % self._vte.get_parent_window().xid)
 
-    dbg ('SEGBUG: Forking command: "%s" with args "%s", loglastlog = "%s", logwtmp = "%s", logutmp = "%s" and cwd "%s"'%(shell, args, login, update_records, update_records, self.cwd))
-    self._pid = self._vte.fork_command (command = shell, argv = args, envv = [], loglastlog = login, logwtmp = update_records, logutmp = update_records, directory=self.cwd)
+    dbg ('SEGBUG: Forking command: "%s" with args "%s", loglastlog = "%s", ' \
+        'logwtmp = "%s", logutmp = "%s" and cwd "%s"' % (shell, args, login,
+            update_records, update_records, self.cwd))
+    self._pid = self._vte.fork_command (command = shell, argv = args,
+        envv = [], loglastlog = login, logwtmp = update_records,
+        logutmp = update_records, directory=self.cwd)
+
+    dbg ('SEGBUG: Forked command')
+
+    self.on_vte_title_change(self._vte) # Force an initial update of our titles
 
-    dbg ('SEGBUG: Forked command') 
     if self._pid == -1:
       err (_('Unable to start shell: ') + shell)
       return (-1)
@@ -416,7 +472,7 @@ text/plain
     """ Return the current working directory of the subprocess.
         This function requires OS specific behaviours
     """
-    cwd = self.pid_get_cwd (self._pid)
+    cwd = self.terminator.pid_get_cwd (self._pid)
     dbg ('get_cwd found: %s'%cwd)
     return (cwd)
 
@@ -453,10 +509,11 @@ text/plain
     self._vte.set_delete_binding (delbind)
 
     # Set our font
-    try:
-      self._vte.set_font (pango.FontDescription (self.conf.font))
-    except:
-      pass
+    if not self._custom_font_size:
+      try:
+        self._vte.set_font (pango.FontDescription (self.conf.font))
+      except:
+        pass
 
     # Set our boldness
     self._vte.set_allow_bold (self.conf.allow_bold)
@@ -490,21 +547,24 @@ text/plain
       self._vte.set_scroll_background(False)
 
     # set transparency for the background (image)
+    opacity = 65535
     if background_type in ("image", "transparent"):
       self._vte.set_background_tint_color (bg_color)
       self._vte.set_background_saturation(1 - (self.conf.background_darkness))
-      self._vte.set_opacity(int(self.conf.background_darkness * 65535))
+      opacity = int(self.conf.background_darkness * 65535)
     else:
       self._vte.set_background_saturation(1)
-      self._vte.set_opacity(65535)
+      
+    if self._composited_support:
+      self._vte.set_opacity(opacity)
 
-    if not self._vte.is_composited():
+    if self._composited_support and not self._vte.is_composited():
       self._vte.set_background_transparent (background_type == "transparent")
     else:
       self._vte.set_background_transparent (False)
 
     # Set our cursor blinkiness
-    self._vte.set_cursor_blinks = (self.conf.cursor_blink)
+    self._vte.set_cursor_blinks (self.conf.cursor_blink)
 
     # Set our audible belliness
     silent_bell = self.conf.silent_bell
@@ -558,7 +618,6 @@ text/plain
       return False
 
     # Left mouse button should transfer focus to this vte widget
-    #LP#242612:
     # we also need to give focus on the widget where the paste occured
     if event.button in (1 ,2):
       self._vte.grab_focus ()
@@ -566,7 +625,7 @@ text/plain
 
     # Right mouse button should display a context menu if ctrl not pressed
     if event.button == 3 and event.state & gtk.gdk.CONTROL_MASK == 0:
-      self.do_popup (event)
+      self.create_popup_menu (self._vte, event)
       return True
 
   def on_vte_notify_enter (self, term, event):
@@ -595,100 +654,97 @@ text/plain
 
   #keybindings for the individual splited terminals (affects only the
   #the selected terminal)
+  UnhandledKeybindings = ('close_window', 'full_screen')
   def on_vte_key_press (self, term, event):
-    keyname = gtk.gdk.keyval_name (event.keyval)
+    mapping = self.terminator.keybindings.lookup(event)
 
-    mask = gtk.gdk.CONTROL_MASK
-    if (event.state & mask) == mask:
-      if keyname == 'plus':
-        self.zoom (True)
-        return (True)
-      elif keyname == 'minus':
-        self.zoom (False)
-        return (True)
-      elif keyname == 'equal':
-        self.zoom_orig ()
-        return (True)
+    if mapping and mapping not in self.UnhandledKeybindings:
+      dbg("on_vte_key_press: lookup found %s" % repr(mapping))
+      getattr(self, "key_" + mapping)()
+      return True
 
-    mask = gtk.gdk.CONTROL_MASK | gtk.gdk.SHIFT_MASK | gtk.gdk.MOD1_MASK
-    if (event.state & mask) == mask:
-      #Top level tab
-      if keyname == 'T':
-        self.terminator.newtab (self, True)
-        return (True)
-    # bindings that should be moved to Terminator as they all just call
-    # a function of Terminator. It would be cleaner is TerminatorTerm
-    # has absolutely no reference to Terminator.
-    # N (next) - P (previous) - O (horizontal) - E (vertical) - W (close)
+    return False
 
-    mask = gtk.gdk.CONTROL_MASK | gtk.gdk.SHIFT_MASK
-    if (event.state & mask) == mask:
-      if keyname == 'N':
-        self.terminator.go_next (self)
-        return (True)
-      elif keyname == "P":
-        self.terminator.go_prev (self)
-        return (True)
-      elif keyname == 'O':
-        self.terminator.splitaxis (self, False)
-        return (True)
-      elif keyname == 'E':
-        self.terminator.splitaxis (self, True)
-        return (True)
-      elif keyname == 'W':
-        self.terminator.closeterm (self)
-        return (True)
-      elif keyname == 'C':
-        self._vte.copy_clipboard ()
-        return (True)
-      elif keyname == 'V':
-        self.paste_clipboard ()
-        return (True)
-      elif keyname == 'S':
-        self.do_scrollbar_toggle ()
-        return (True)
-      elif keyname == 'T':
-        self.terminator.newtab(self)
-        return (True)
-      elif keyname in ('Up', 'Down', 'Left', 'Right'):
-          self.terminator.resizeterm (self, keyname)
-          return (True)
-      elif keyname  == 'Page_Down':
-          self.terminator.move_tab(self, 'right')
-          return (True)
-      elif keyname == 'Page_Up':
-          self.terminator.move_tab(self, 'left')
-          return (True)
-      elif keyname == 'Z':
-        self.terminator.toggle_zoom (self, True)
-        return (True)
-      elif keyname == 'X':
-        self.terminator.toggle_zoom (self)
-        return (True)
-      
-    mask = gtk.gdk.CONTROL_MASK
-    if (event.state & mask) == mask:
-      if keyname  == 'Page_Down':
-          self.terminator.next_tab(self)
-          return (True)
-      elif keyname == 'Page_Up':
-          self.terminator.previous_tab(self)
-          return (True)
-    
-    if keyname and (keyname == 'Tab' or keyname.endswith('_Tab')):
-        mask = gtk.gdk.CONTROL_MASK | gtk.gdk.SHIFT_MASK
-        if (event.state & mask) == mask:
-            self.terminator.go_prev (self)
-            return (True)
-        mask = gtk.gdk.CONTROL_MASK
-        if (event.state & mask) == mask:
-            self.terminator.go_next (self)
-            return (True)
-    # Warning, mask value is either gtk.gdk.CONTROL_MASK or gtk.gdk.CONTROL_MASK | gtk.gdk.SHIFT_MASK
-    # if you intend to use it, reinit it
-    return (False)
+  # Key events
+  def key_zoom_in(self):
+    self.zoom (True)
+
+  def key_zoom_out(self):
+    self.zoom (False)
+
+  def key_copy(self):
+    self._vte.copy_clipboard ()
+
+  def key_paste(self):
+    self.paste_clipboard ()
+
+  def key_toggle_scrollbar(self):
+    self.do_scrollbar_toggle ()
+
+  def key_zoom_normal(self):
+    self.zoom_orig ()
+
+  def key_search(self):
+    self.start_search()
+
+  # bindings that should be moved to Terminator as they all just call
+  # a function of Terminator. It would be cleaner if TerminatorTerm
+  # has absolutely no reference to Terminator.
+  # N (next) - P (previous) - O (horizontal) - E (vertical) - W (close)
+  def key_new_root_tab(self):
+    self.terminator.newtab (self, True)
+
+  def key_go_next(self):
+    self.terminator.go_next (self)
+
+  def key_go_prev(self):
+    self.terminator.go_prev (self)
+
+  def key_split_horiz(self):
+    self.terminator.splitaxis (self, False)
+
+  def key_split_vert(self):
+    self.terminator.splitaxis (self, True)
+
+  def key_close_term(self):
+    self.terminator.closeterm (self)
+
+  def key_new_tab(self):
+    self.terminator.newtab(self)
+
+  def key_resize_up(self):
+    self.terminator.resizeterm (self, 'Up')
+
+  def key_resize_down(self):
+    self.terminator.resizeterm (self, 'Down')
+
+  def key_resize_left(self):
+    self.terminator.resizeterm (self, 'Left')
+
+  def key_resize_right(self):
+    self.terminator.resizeterm (self, 'Right')
+
+  def key_move_tab_right(self):
+    self.terminator.move_tab (self, 'right')
+
+  def key_move_tab_left(self):
+    self.terminator.move_tab (self, 'left')
+
+  def key_toggle_zoom(self):
+    self.terminator.toggle_zoom (self)
+
+  def key_scaled_zoom(self):
+    self.terminator.toggle_zoom (self, True)
+
+  def key_next_tab(self):
+    self.terminator.next_tab (self)
+
+  def key_prev_tab(self):
+    self.terminator.previous_tab (self)
+  # End key events
 
   def zoom_orig (self):
+    self._custom_font_size = None
     self._vte.set_font (pango.FontDescription (self.conf.font))
 
   def zoom (self, zoom_in):
@@ -701,25 +757,90 @@ text/plain
       fontsize += pango.SCALE
 
     pangodesc.set_size (fontsize)
+    self._custom_font_size = fontsize
     self._vte.set_font (pangodesc)
 
-  def on_vte_popup_menu (self, term, event):
-    self.do_popup (event)
+  def start_search(self):
+    self._searchbox.show()
+    self._searchinput.grab_focus()
 
-  def do_popup (self, event = None):
-    menu = self.create_popup_menu (event)
-    menu.popup (None, None, None, event.button, event.time)
+  def search_keypress(self, widget, event):
+    key = gtk.gdk.keyval_name(event.keyval)
+    if key == 'Escape':
+      self.end_search()
 
-  def create_popup_menu (self, event):
+  def end_search(self):
+    self._search_row = 0
+    self._search_string = None
+    self._search_result_label.set_text("")
+    self._searchbox.hide()
+    self._scrollbar.set_value(self._vte.get_cursor_position()[1])
+    self._vte.grab_focus()
+
+  def do_search(self, widget):
+    string = widget.get_text()
+    dbg("do_search: Looking for %s" % repr(string))
+    if string == '':
+      return
+
+    if string != self._search_string:
+      self._search_row = self._get_vte_buffer_range()[0]
+      self._search_string = string
+
+    self._search_result_label.set_text("Searching scrollback")
+    self.next_search()
+
+  # Called by get_text_range, once per character.  Argh.
+  def _search_character(self, widget, col, row, junk):
+    return True
+
+  def next_search(self):
+    startrow,endrow = self._get_vte_buffer_range()
+    while True:
+      if self._search_row == endrow:
+        self._search_row = startrow
+        self._search_result_label.set_text("Finished Search")
+        return
+      buffer = self._vte.get_text_range(self._search_row, 0, self._search_row, -1, self._search_character)
+
+      # dbg("Row %d buffer: %s" % (self._search_row, repr(buffer)))
+      index = buffer.find(self._search_string)
+      if index != -1:
+        self._search_result_label.set_text("Found at row %d" % self._search_row)
+        self._scrollbar.set_value(self._search_row)
+        self._search_row += 1
+        return
+      self._search_row += 1
+
+  def _get_vte_buffer_range(self):
+    column, endrow = self._vte.get_cursor_position()
+    startrow = max(0, endrow - self.conf.scrollback_lines)
+    return(startrow, endrow)
+
+  def create_popup_menu (self, widget, event = None):
     menu = gtk.Menu ()
     url = None
 
     if event:
       url = self._vte.match_check (int (event.x / self._vte.get_char_width ()), int (event.y / self._vte.get_char_height ()))
+      button = event.button
+      time = event.time
+    else:
+      button = 0
+      time = 0
 
     if url:
       if url[1] != self.matches['email']:
-        address = url[0]
+        # Add protocol if we launch a URL without it, otherwise xdg-open won't open it
+        if url[1] == self.matches['addr_only']:
+          if url[0][0:3] == "ftp":
+              # "ftp.foo.bar" -> "ftp://ftp.foo.bar"
+              address = "ftp://" + url[0]
+          else:
+              # Assume http
+              address = "http://" + url[0]
+        else:
+          address = url[0]
         nameopen = _("_Open Link")
         namecopy = _("_Copy Link Address")
         iconopen = gtk.image_new_from_stock(gtk.STOCK_JUMP_TO, gtk.ICON_SIZE_MENU)
@@ -784,7 +905,6 @@ text/plain
 
       item.connect ("activate", lambda menu_item: self.terminator.splitaxis (self, False))
       menu.append (item)
-
       item = gtk.ImageMenuItem (str_vert)
       item_image = gtk.Image ()
       item_image.set_from_icon_name (APP_NAME + '_vert', gtk.ICON_SIZE_MENU)
@@ -792,11 +912,17 @@ text/plain
 
       item.connect ("activate", lambda menu_item: self.terminator.splitaxis (self, True))
       menu.append (item)
-    
+
       item = gtk.MenuItem (_("Open _Tab"))
       item.connect ("activate", lambda menu_item: self.terminator.newtab (self))
       menu.append (item)
 
+      if self.terminator.debugaddress:
+        item = gtk.MenuItem (_("Open _Debug Tab"))
+        item.connect ("activate", lambda menu_item: self.terminator.newtab (self, command = "telnet %s" % ' '.join([str(x) for x in self.terminator.debugaddress])))
+        menu.append (item)
+
+
       if self.conf.extreme_tabs:
         item = gtk.MenuItem (_("Open Top Level Tab"))
         item.connect ("activate", lambda menu_item: self.terminator.newtab (self, True))
@@ -811,7 +937,7 @@ text/plain
         item.connect ("activate", lambda menu_item: self.terminator.toggle_zoom (self, True))
         menu.append (item)
 
-        item = gtk.MenuItem (_("_Maximise terminal"))
+        item = gtk.MenuItem (_("Ma_ximise terminal"))
         item.connect ("activate", lambda menu_item: self.terminator.toggle_zoom (self))
         menu.append (item)
       else:
@@ -821,7 +947,7 @@ text/plain
           menu.append (item)
 
         if self.terminator._zoomed and self.terminator._maximised:
-          item = gtk.MenuItem (_("U_nmaximise terminal"))
+          item = gtk.MenuItem (_("Unma_ximise terminal"))
           item.connect ("activate", lambda menu_item: self.terminator.toggle_zoom (self))
           menu.append (item)
 
@@ -833,7 +959,9 @@ text/plain
     menu.append (item)
 
     menu.show_all ()
-    return menu
+    menu.popup (None, None, None, button, time)
+
+    return True
 
   def on_encoding_change (self, widget, encoding):
     current = self._vte.get_encoding ()
@@ -889,19 +1017,28 @@ text/plain
       
       radioitem.connect ('activate', self.on_encoding_change, encoding[1])
       submenu.append (radioitem)
-   
+
+  def get_window_title(self, vte = None):
+    if vte is None:
+      vte = self._vte
+    title = vte.get_window_title ()
+    if title is None:
+      title = str(self.command)
+    return title
+
   def on_vte_title_change(self, vte):
+    title = self.get_window_title(vte)
     if self.conf.titletips:
       vte.set_property ("has-tooltip", True)
-      vte.set_property ("tooltip-text", vte.get_window_title ())
+      vte.set_property ("tooltip-text", title)
     #set the title anyhow, titlebars setting only show/hide the label
-    self._title.set_text(vte.get_window_title ())
-    self.terminator.set_window_title("%s: %s" %(APP_NAME.capitalize(), vte.get_window_title ()))
+    self._title.set_text(title)
+    self.terminator.set_window_title("%s - %s" % (title, APP_NAME.capitalize()))
     notebookpage = self.terminator.get_first_notebook_page(vte)
     while notebookpage != None:
       if notebookpage[0].get_tab_label(notebookpage[1]):
         label = notebookpage[0].get_tab_label(notebookpage[1])
-        label.set_title(vte.get_window_title ())
+        label.set_title(title)
         notebookpage[0].set_tab_label(notebookpage[1], label)
       notebookpage = self.terminator.get_first_notebook_page(notebookpage[0])
 
@@ -909,22 +1046,22 @@ text/plain
     self._titlebox.modify_bg(gtk.STATE_NORMAL,self.terminator.window.get_style().bg[gtk.STATE_SELECTED])
     self._title.modify_fg(gtk.STATE_NORMAL, self.terminator.window.get_style().fg[gtk.STATE_SELECTED])
     return
-    
+
   def on_vte_focus_out(self, vte, event):
     self._titlebox.modify_bg(gtk.STATE_NORMAL, self.terminator.window.get_style().bg[gtk.STATE_NORMAL])
     self._title.modify_fg(gtk.STATE_NORMAL, self.terminator.window.get_style().fg[gtk.STATE_NORMAL])
     return
 
   def on_vte_focus(self, vte):
-    if vte.get_window_title ():
-      self.terminator.set_window_title("%s: %s" %(APP_NAME.capitalize(), vte.get_window_title ()))
-      notebookpage = self.terminator.get_first_notebook_page(vte)
-      while notebookpage != None:
-        if notebookpage[0].get_tab_label(notebookpage[1]):
-          label = notebookpage[0].get_tab_label(notebookpage[1])
-          label.set_title(vte.get_window_title ())
-          notebookpage[0].set_tab_label(notebookpage[1], label)
-        notebookpage = self.terminator.get_first_notebook_page(notebookpage[0])
+    title = self.get_window_title(vte)
+    self.terminator.set_window_title("%s - %s" % (title, APP_NAME.capitalize()))
+    notebookpage = self.terminator.get_first_notebook_page(vte)
+    while notebookpage != None:
+      if notebookpage[0].get_tab_label(notebookpage[1]):
+        label = notebookpage[0].get_tab_label(notebookpage[1])
+        label.set_title(title)
+        notebookpage[0].set_tab_label(notebookpage[1], label)
+      notebookpage = self.terminator.get_first_notebook_page(notebookpage[0])
 
   def destroy(self):
     self._vte.destroy()
